from typing import Any, Callable, List, Tuple


class ModuleSpec:
    """Specification for a executable module.

    Args:
        name:
            Name of the module
        entrypoint:
            Fn to be called as an entry to the module. Entrypoint receives
            args intended for the module and should return tuple of output and
            Exception occured.
    """
    def __init__(
        self,
        name: str,
        entrypoint: Callable[[Tuple[str, ...]], Any]
    ) -> None:
        self._name = name
        self._entrypoint = entrypoint

    @property
    def name(self) -> str:
        return self._name

    @property
    def entrypoint(self):
        return self._entrypoint



from functools import partial
import sys
from typing import Any, Callable, List, Tuple, Type

from loguru import logger
from clyjin.core.ArgsTuple import ArgsTuple
from clyjin.core.BUILTIN_MODULE_SPECS import BUILTIN_MODULE_SPECS
from clyjin.core.ModuleSpec import ModuleSpec
from clyjin.core.log import init_logger

class CannotFindSpecException(Exception): pass

class CliArgs:
    """Parsed CLI args.
    
    Args:
        module_name:
            Name of the module spec targets
        args:
            List of arguments for this module
    """
    def __init__(
        self,
        module_name: str,
        args: ArgsTuple
    ) -> None:
        self._module_name = module_name
        self._args = args

    @property
    def module_name(self):
        return self._module_name

    @property
    def args(self):
        return self._args

class NoModuleNameException(Exception): pass

def main() -> None:
    init_logger()

    args: CliArgs = _parse_args(tuple(sys.argv)) 
    _process(args)

def _parse_args(args: ArgsTuple) -> CliArgs:
    try:
        module_name: str = args[1]
    except IndexError as err:
        raise NoModuleNameException()
    else:
        return CliArgs(module_name=module_name, args=args[2:])

def _process(args: CliArgs) -> None:
    spec = _pick_spec(args.module_name)
    spec.entrypoint(args.args)

def _pick_spec(module_name: str) -> ModuleSpec:
    for spec in BUILTIN_MODULE_SPECS:
        if module_name == spec.name:
            return spec
    raise CannotFindSpecException()



"""Create flask application hosts OpenAPI documentation."""
import os
from functools import reduce
from typing import Any, Generator, Iterator, List, Sequence, Tuple
import aiohttp_jinja2
import jinja2
from loguru import logger
from aiohttp import web

import yaml
from clyjin.core.ArgsTuple import ArgsTuple

BIND_FLAG_OPTIONS = ["-b", "--bind"]
API_PATHS_FILE_FLAG_OPTIONS = ["-f", "--file"]


def _load_api_paths(path: str) -> Iterator[str]:
    with open(path, "r") as f:
        paths: List[str] = f.readlines()
    yield from map(
        lambda path: path.strip(),
        filter(lambda path: not path.isspace(), paths)
    )

def _get_api_name_from_entrypoint(path: str) -> str:
    return yaml.load(open(path, "r"), yaml.SafeLoader)["info"]["title"]

def _parse_api_entrypoint_path_by_name(
        name: str,
        args: ArgsTuple
    ) -> Iterator[str]:
    # Find appropriate file arg option, load file for this option, create spec
    # out of this file content specified paths
    api_paths = _load_api_paths(args[
            args.index(
                next(filter(
                    lambda option: option in args,
                    API_PATHS_FILE_FLAG_OPTIONS
                ))
            ) + 1
        ]
    )
    yield from filter(
        # Lower both sides for better matching of capitalized packages, i.e.
        # in routes they will always be responsive for any-cased variants
        lambda path:
            _get_api_name_from_entrypoint(path).lower() == name.lower(),
        api_paths
    )

def _validate_api_dirname(
        dirname: str,
        args: ArgsTuple
    ) -> None:
    # FIXME:
    #   If dirname will be a folder inside actual correct dirname, e.g.
    #   .../oapi/player/player.yml - it won't validate this - need to implement
    #   recursion to check all the way until match occurs or top of the tree is
    #   reached

    next(
        filter(
            lambda path: os.path.dirname(path) == dirname,
            _load_api_paths(args[
                    args.index(
                        next(filter(
                            lambda option: option in args,
                            API_PATHS_FILE_FLAG_OPTIONS
                        ))
                    ) + 1
                ]
            )
        )
    )

def _parse_binded_host_port(args: ArgsTuple) -> Tuple[str, int]:
    def _parse(bind_string: str) -> Tuple[str, int]:
        host, port =  bind_string.split(":")
        port = int(port)
        return host, port

    for option in BIND_FLAG_OPTIONS:
        if option in args:
            return _parse(args[args.index(option)+1])

    return "0.0.0.0", 5000

def main(args: Tuple[str]) -> None:
    async def _get_public(req):
        # FIXME:
        #   Security problems, rewrite or add additional checks to disallow path
        #   building. But if i use path as a route var, seems like that browser
        #   disallows such types of routes.
        path = str(req.match_info["path"])
        return web.FileResponse(
            os.path.join(
                "clyjin/modules/api/public",
                path
            )
        )

    async def _get_oapi(req):
        module_name = str(req.match_info["module_name"])

        entrypoint_path = next(
            _parse_api_entrypoint_path_by_name(module_name, args)
        )

        api_url = "/raw/{}".format(
            entrypoint_path
        )

        return aiohttp_jinja2.render_template(
            "index.html",
            req,
            {"api_entrypoint_url": api_url}
        )

    async def _get_raw(req):
        path = str(req.match_info["path"])

        # Spec by dirname should be found to disallow random file exposing
        _validate_api_dirname(
            os.path.dirname(path),
            args
        )

        return web.FileResponse(
            path
        )

    #-- Initialization --#
    app = web.Application()
    aiohttp_jinja2.setup(app,
        loader=jinja2.FileSystemLoader("clyjin/modules/api/templates"))

    app.add_routes([
        web.get(r"/public/{path:.*}", _get_public),
        web.get(r"/raw/{path:.*}", _get_raw),
        web.get(r"/{module_name}", _get_oapi),
    ])

    host, port = _parse_binded_host_port(args)
    web.run_app(app, host=host, port=port)

    # @app.route("/<api_name>")
    # def swagger(api_name: str):

    #     return render_template(
    #         "index.html",
    #         api_entrypoint_url=api_url
    #     )

    # app.run(host=args.host, port=args.port)



import typing

class UnconvertableStringException(Exception):
    def __init__(self, string: str, *args: object) -> None:
        message = "Unconvertable string {}".format(string)
        super().__init__(message, *args)

def convert_string_to_bool(string: str) -> bool:
    """Convert given string to bool according to logical rules.
    
    Possible values of input string:
        - "true"
        - "false"
        - "0"
        - "1"

    Args:
        string:
            String to be converted

    Return:
        Converted boolean

    Raise:
        UnconvertableStringException
    """
    if string in ["true", "1"]:
        return True
    elif string in ["false", "0"]:
        return False
    else:
        raise UnconvertableStringException(string)


import os
import sys
import traceback
from typing import Any, Tuple, Union

from clyjin.core import convert_string_to_bool
from loguru import logger


class UnrecognizedLevelException(Exception):
    def __init__(self, level: str, *args) -> None:
        message = "Unrecognized level {}".format(level)
        super().__init__(message, *args)

class ExceptionsOnlyAllowedException(Exception):
    def __init__(self, obj: object, *args: object) -> None:
        message = "Object {} is not an Exception".format(obj)
        super().__init__(message, *args)

def init_logger() -> None:
    """Initialize logger"""
    SINK = os.environ.get("CLYJIN_LOG_SINK", "var/logs/app.log")
    # LEVEL = os.environ.get("CLYJIN_LOG_LEVEL", "INFO")
    LEVEL = os.environ.get("CLYJIN_LOG_LEVEL", "DEBUG")
    HAS_TO_SERIALIZE: bool = convert_string_to_bool(
        os.environ.get("CLYJIN_HAS_TO_SERIALIZE", "true")
    )

    logger.add(
        SINK,
        level=LEVEL,
        serialize=HAS_TO_SERIALIZE
    )

# DEPRECATED:
#   Custom formatting no more used... builtin loguru formatter is
#   better :-)
###
# def _format_log(record: Any) -> str:
#     logged_err: Union[Exception, None] = None
#     src_args = [record["name"], record["function"]]

#     src = ".".join(src_args)

#     level = _colorize_level(record["level"].name)

#     message = record["message"] if not logged_err else " ".join(
#         logged_err.args
#     )
#     # Reformat src and messages related to module to avoid ANSI errors
#     src = src.replace("<module>", "__module__")
#     message = message.replace("<module>", "__module__")

#     # Time is not required by now, but here is the method for its formatting
#     # print(record["time"].strftime("%Y.%m.%d at %H:%M:%S:%f%z"))

#     # Add traceback info if exception is presented
#     if record.get("exception", None):
#         message += "\n" + traceback.format_exc()

#     return "[{src} / {level}] {message}\n".format(
#         src=src, level=level, message=message
#     )

# def _colorize_level(level: str) -> str:
#     wrapped: str

#     if level == "DEBUG":
#         wrapped = "<magenta>{}</magenta>"
#     elif level == "INFO":
#         wrapped = "<blue>{}</blue>"
#     elif level == "WARNING":
#         wrapped = "<yellow>{}</yellow>"
#     elif level == "ERROR":
#         wrapped = "<red>{}</red>"
#     elif level == "CRITICAL":
#         wrapped = "<black>{}</black>"
#     else:
#         raise UnrecognizedLevelException(level)

#     return wrapped.format(level)
###



from typing import List
from clyjin.core.ModuleSpec import ModuleSpec
from clyjin.modules import api, boot

BUILTIN_MODULE_SPECS: List[ModuleSpec] = [
    ModuleSpec("boot", boot.main),
    ModuleSpec("api", api.main)
] 


from setuptools import setup, find_packages

from clyjin import __version__ as version


with open("requirements.txt", "r") as file:
    install_requires = [x.strip() for x in file.readlines()]

with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="clyjin",
    version=version,
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    description="System configuration toolbox",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Alexander Ryzhov",
    author_email="thed4rkof@gmail.com",
    url="https://github.com/ryzhovalex/clyjin",
    keywords=["cli", "automatization", "system", "os"],
    entry_points={
        "console_scripts": [
            "clyjin = clyjin.core.cli:main",
            "cj = clyjin.core.cli:main"
        ],
    },
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 1 - Planning",

        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",

        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.7",
    ],
)
